from .env_and_links import make_mad_environment
from .madx_model import attach_beam_to_sequence, configure_b4_from_b2
from .madx_model import save_lines_for_closed_orbit_reference
from .tuning import machine_tuning
from . import yaml


import pandas as pd
import numpy as np

import xtrack as xt

def attach_beam_to_sequence(sequence, beam_to_configure=1, beam_configuration=None):
    """Attach beam to sequence

    Parameters
    ----------
    mad : pymadx.madx.Madx
        Madx object
    beam_to_configure : int
        Beam number to configure
    configuration : dict
        Beam configuration

    Returns
    -------
    None

    """
    mad = sequence._madx

    # beam energy
    mad.globals.nrj = beam_configuration['beam_energy_tot']
    particle_type = 'proton'

    if 'particle_mass' in beam_configuration.keys():
        particle_mass = beam_configuration['particle_mass']
        particle_type = 'ion'
    else:
        particle_mass = mad.globals.pmass # proton mass

    if 'particle_charge' in beam_configuration.keys():
        particle_charge = beam_configuration['particle_charge']
        particle_type = 'ion'
    else:
        particle_charge = 1.

    gamma_rel = (beam_configuration['beam_energy_tot'])/particle_mass
    # bv and bv_aux flags
    if beam_to_configure == 1:
        ss_beam_bv, ss_bv_aux = 1, 1
    elif beam_to_configure == 2:
        ss_beam_bv, ss_bv_aux = -1, 1
    elif beam_to_configure == 4:
        ss_beam_bv, ss_bv_aux = 1, -1
    else:
        raise ValueError("beam_to_configure must be 1, 2 or 4")

    mad.globals['bv_aux'] = ss_bv_aux
    mad.input(f'''
    beam, particle={particle_type},sequence={sequence.name},
        energy={beam_configuration['beam_energy_tot']},
        sigt={beam_configuration.get('beam_sigt', 0.0001)},
        bv={ss_beam_bv},
        npart={beam_configuration.get('beam_npart', 1)},
        sige={beam_configuration.get('beam_sige', 1e-6)},
        ex={beam_configuration.get('beam_norm_emit_x', 1) * 1e-6 / gamma_rel},
        ey={beam_configuration.get('beam_norm_emit_y', 1) * 1e-6 / gamma_rel},
        mass={particle_mass},
        charge={particle_charge};
    ''')

    # Used for the rescaling of spectrometers
    mad.use(sequence.name)
    mad.globals.nrj = sequence.beam.energy / sequence.beam.charge

def save_lines_for_closed_orbit_reference(sequence_clockwise, sequence_anticlockwise):

    lines_co_ref = {}
    if sequence_clockwise is not None:
        name_cw = sequence_clockwise.name
        lines_co_ref[name_cw + '_co_ref'] = xt.Line.from_madx_sequence(
            sequence_clockwise,
            deferred_expressions=True,
            expressions_for_element_types=('kicker', 'hkicker', 'vkicker'),
            replace_in_expr={'bv_aux': 'bvaux_' + name_cw})

    if sequence_anticlockwise is not None:
        name_acw = sequence_anticlockwise.name
        lines_co_ref[name_acw + '_co_ref'] = xt.Line.from_madx_sequence(
            sequence_anticlockwise,
            deferred_expressions=True,
            expressions_for_element_types=('kicker', 'hkicker', 'vkicker'),
            replace_in_expr={'bv_aux': 'bvaux_' + name_acw})

    return lines_co_ref



def configure_b4_from_b2(sequence_b4, sequence_b2,
        update_globals={'bv_aux': -1, 'mylhcbeam': 4}):

    mad_b2 = sequence_b2._madx
    mad_b4 = sequence_b4._madx

    var_dicts_b2 = _get_variables_dicts(mad_b2)
    var_dicts_b4 = _get_variables_dicts(mad_b4)

    b2_const=var_dicts_b2['constants']
    b4_const=var_dicts_b4['constants']
    for nn in b2_const.keys():
        if nn[0]=='_':
            print(f'The constant {nn} cannot be assigned!')
        else:
            if nn not in b4_const.keys():
                mad_b4.input(f'const {nn}={b2_const[nn]:.50e}')

    b2_indep=var_dicts_b2['independent_variables']
    b4_indep=var_dicts_b4['independent_variables']
    for nn in b2_indep.keys():
        mad_b4.input(f'{nn}={b2_indep[nn]:.50e}')

    b2_dep=var_dicts_b2['dependent_variables_expr']
    b4_dep=var_dicts_b4['dependent_variables_expr']
    for nn in b2_dep.keys():
        mad_b4.input(f'{nn}:={str(b2_dep[nn])}')

    # bv_aux and my my lhcbeam need to be defined explicitly
    for nn in update_globals.keys():
        mad_b4.input(f'{nn}={update_globals[nn]}')

    # Attach beam
    mad_b4.use(sequence_b2.name)
    beam_command = str(sequence_b2.beam)
    assert(', bv=-1.0' in beam_command)
    beam_command = beam_command.replace(', bv=-1.0', ', bv=1.0')
    mad_b4.input(beam_command)
    mad_b4.use(sequence_b4.name)

    # CHECKS
    var_dicts_b2 = _get_variables_dicts(mad_b2)
    var_dicts_b4 = _get_variables_dicts(mad_b4)

    b2_const=var_dicts_b2['constants']
    b4_const=var_dicts_b4['constants']
    for nn in b4_const.keys():
        assert b2_const[nn] == b4_const[nn]

    for nn in b2_const.keys():
        if nn not in b4_const.keys():
            print(f'Warning: b2 const {nn}={b2_const[nn]} is not in b4.')

    b2_indep=var_dicts_b2['independent_variables']
    b4_indep=var_dicts_b4['independent_variables']
    for nn in b2_indep.keys():
        if str(nn) in list(update_globals.keys()):
            continue
        assert b4_indep[nn] == b2_indep[nn]

    for nn in b4_indep.keys():
        if nn not in b2_indep.keys():
            print(f'Warning: b4 indep {nn}={b4_indep[nn]} is not in b2.')

    b2_dep=var_dicts_b2['dependent_variables_expr']
    b4_dep=var_dicts_b4['dependent_variables_expr']
    for nn in b2_dep.keys():
        if str(nn) in list(update_globals.keys()):
            continue
        assert str(b4_dep[nn]) == str(b2_dep[nn])

    for nn in b4_dep.keys():
        if nn not in b2_dep.keys():
            print(f'Warning: b4 dep {nn}={str(b4_dep[nn])} is not in b2.')

def _get_variables_dicts(mad, expressions_as_str=True):
        variables_df = _get_variables_dataframes(mad)
        outp = {
            'constants': variables_df['constants'].to_dict()['value'],
            'independent_variables':
                variables_df['independent_variables'].to_dict()['value'],
            'dependent_variables_expr':
                variables_df['dependent_variables'].to_dict()['expression'],
            'dependent_variables_val':
                variables_df['dependent_variables'].to_dict()['value'],
            }

        outp['all_variables_val'] = {kk:outp['constants'][kk] for
                kk in outp['constants'].keys()}
        outp['all_variables_val'].update(outp['independent_variables'])
        outp['all_variables_val'].update(outp['dependent_variables_val'])


        return outp

def _get_variables_dataframes(mad, expressions_as_str=True):
    '''
    Extract the dictionary of the variables and constant pandas DF of the -X global workspace.

    Returns:
        The a dictionary containing:
        - constants_df: the pandas DF with the constants
        - independent_variables: the pandas DF with the independent variables
        - dependent_variables: the pandas DF with the dependent variables
    All the three DFs have a columns 'value' with the numerical values of the costants/variables.
    The dependent_variable_df, in addition to 'value' has the following columns:
        - 'expression': the string corrensponding to the MAD-X expression
        - 'parameters': the list of parameters used in the expression
        - 'knobs': the list of the independent variables that control
            the dependent variables. Note tha the parameters can be constants and/or dependent variables,
            whereas the 'knobs' are only independent variables.
    '''
    my_dict={}
    aux=_independent_variables_df(mad)
    import numpy as np
    independent_variables_df=aux[np.logical_not(aux['constant'])].copy()
    del independent_variables_df['constant']
    constant_df=aux[aux['constant']].copy()
    del constant_df['constant']
    my_dict['constants']=constant_df
    my_dict['independent_variables']=independent_variables_df
    my_dict['dependent_variables']=_dependent_variables_df(mad)

    if expressions_as_str:
        my_dict['dependent_variables']['expression'] = (
                    my_dict['dependent_variables']['expression'].apply(
                        str))
    return my_dict

def _independent_variables_df(mad):
    '''
    Extract the pandas DF with the independent variables of the MAD-X handle.

    Returns:
        The pandas DF of the independent variables. The columns of the DF correspond to the
        - the numerical value of the independent variable (value)
        - a boolean value to know it the variable is constant or not (constant)

    See madxp/examples/variablesExamples/000_run.py
    '''

    dep_df=_dependent_variables_df(mad)
    #if len(dep_df)>0:
    #    aux=list(dep_df['knobs'].values)
    #    aux=list(itertools.chain.from_iterable(aux))
    #fundamentalSet=set(np.unique(aux))
    independent_variable_set=set(mad.globals)-set(dep_df.index)
    my_dict={}
    for i in independent_variable_set:
        my_dict[i]={}
        if mad._libmadx.get_var_type(i)==0:
            my_dict[i]['constant']=True
            # my_dict[i]['knob']=False
        else:
            my_dict[i]['constant']=False
            # if i in fundamentalSet:
            #    my_dict[i]['knob']=True
            # else:
            #    my_dict[i]['knob']=False
        my_dict[i]['value']=mad.globals[i]

    return pd.DataFrame(my_dict).transpose()[['value','constant']].sort_index()

def _dependent_variables_df(mad):
    '''
    Extract the pandas DF with the dependent variables of the MAD-X handle.

    Returns:
        The pandas DF of the dependent variables. The columns of the DF correspond to the
        - the numerical value of the dependent variable (value)
        - the string corresponding to the MAD-X expression (expression)
        - the list of parameters used in the expression (parameters)
        - the list of the fundamental independent variables.
            These are independent variables that control numerical values of the variable (knobs).

    See madxp/examples/variablesExamples/000_run.py
    '''
    my_dict={}
    for i in list(mad.globals):
        aux=_extract_parameters(str(mad._libmadx.get_var(i)))
        if aux!=[]:
            my_dict[i]={}
            my_dict[i]['parameters']=list(np.unique(aux))

    myhash=hash(str(my_dict))
    while True:
        for i in my_dict:
            aux=[]
            for j in my_dict[i]['parameters']:
                try:
                    aux=aux+my_dict[j]['parameters']
                except:
                    aux=aux+[j]
            my_dict[i]['knobs']=list(np.unique(aux))
        if myhash==hash(str(my_dict)):
            break
        else:
            myhash=hash(str(my_dict))

    for i in my_dict:
        for j in my_dict[i]['knobs'].copy():
            if mad._libmadx.get_var_type(j)==0:
                my_dict[i]['knobs'].remove(j)
        my_dict[i]['expression']=mad._libmadx.get_var(i)
        my_dict[i]['value']=mad.globals[i]

    if len(my_dict)>0:
        return pd.DataFrame(my_dict).transpose()[['value','expression','parameters','knobs']].sort_index()
    else:
        return pd.DataFrame()

def _extract_parameters(my_string):
    '''
    Extract all the parameters of a MAD-X expression.
    Args:
        my_string: The string of the MAD-X expression to parse.

    Returns:
        The list of the parameters present in the MAD-X expression.
    '''
    if (type(my_string)=='NoneType' or my_string==None
            or my_string=='None' or my_string=='[None]' or 'table(' in my_string):
        return []
    else:
        for i in [
        '*','->','-','/','+','^','(',')','[',']',',','\'','None']:
            my_string=my_string.replace(i,' ')
        my_list=my_string.split(' ')
        my_list=list(np.unique(my_list))
        if '' in my_list:
            my_list.remove('')
        if type(my_list)=='NoneType':
            my_list=[]
        for i in my_list.copy():
            if i.isdigit() or i[0].isdigit() or i[0]=='.':
                my_list.remove(i)
        my_list=list(set(my_list)-
        set([
            'sqrt',
            'log',
            'log10',
            'exp',
            'sin',
            'cos',
            'tan',
            'asin',
            'acos',
            'atan',
            'sinh',
            'cosh',
            'tanh',
            'sinc',
            'abs',
            'erf',
            'erfc',
            'floor',
            'ceil',
            'round',
            'frac',
            'ranf',
            'gauss',
            'tgauss']))
        return my_list


import os

def make_mad_environment(links):

    for kk in links.keys():
        os.system(f'rm {kk}')
        os.symlink(os.path.abspath(links[kk]), kk)

    # Create empty temp folder
    os.system('rm -r temp; mkdir temp')


from scipy.constants import c as clight

import xtrack as xt

def luminosity_leveling(collider, config_lumi_leveling, config_beambeam):
    for ip_name in config_lumi_leveling.keys():

        print(f'\n --- Leveling in {ip_name} ---')

        config_this_ip = config_lumi_leveling[ip_name]
        bump_range = config_this_ip['bump_range']

        assert config_this_ip['preserve_angles_at_ip'], (
            'Only preserve_angles_at_ip=True is supported for now')
        assert config_this_ip['preserve_bump_closure'], (
            'Only preserve_bump_closure=True is supported for now')

        beta0_b1 = collider.lhcb1.particle_ref.beta0[0]
        f_rev=1/(collider.lhcb1.get_length() /(beta0_b1 * clight))

        targets=[]
        vary=[]

        if 'luminosity' in config_this_ip.keys():
            targets.append(
                xt.TargetLuminosity(
                    ip_name=ip_name, luminosity=config_this_ip['luminosity'], crab=False,
                    tol=0.01 * config_this_ip['luminosity'],
                    f_rev=f_rev, num_colliding_bunches=config_this_ip['num_colliding_bunches'],
                    num_particles_per_bunch=config_beambeam['num_particles_per_bunch'],
                    sigma_z=config_beambeam['sigma_z'],
                    nemitt_x=config_beambeam['nemitt_x'],
                    nemitt_y=config_beambeam['nemitt_y'])
            )
        elif 'separation_in_sigmas' in config_this_ip.keys():
            targets.append(
                xt.TargetSeparation(
                    ip_name=ip_name,
                    separation_norm=config_this_ip['separation_in_sigmas'],
                    tol=1e-4, # in sigmas
                    plane=config_this_ip['plane'],
                    nemitt_x=config_beambeam['nemitt_x'],
                    nemitt_y=config_beambeam['nemitt_y'])
            )
        else:
            raise ValueError('Either `luminosity` or `separation_in_sigmas` must be specified')

        if config_this_ip['impose_separation_orthogonal_to_crossing']:
            targets.append(
                xt.TargetSeparationOrthogonalToCrossing(ip_name='ip8'))
        vary.append(
            xt.VaryList(config_this_ip['knobs'], step=1e-4))

        # Target and knobs to rematch the crossing angles and close the bumps
        for line_name in ['lhcb1', 'lhcb2']:
            targets += [
                # Preserve crossing angle
                xt.TargetList(['px', 'py'], at=ip_name, line=line_name, value='preserve', tol=1e-7, scale=1e3),
                # Close the bumps
                xt.TargetList(['x', 'y'], at=bump_range[line_name][-1], line=line_name, value='preserve', tol=1e-5, scale=1),
                xt.TargetList(['px', 'py'], at=bump_range[line_name][-1], line=line_name, value='preserve', tol=1e-5, scale=1e3),
            ]

        vary.append(xt.VaryList(config_this_ip['corrector_knob_names'], step=1e-7))

        # Match
        collider.match(
            lines=['lhcb1', 'lhcb2'],
            ele_start=[bump_range['lhcb1'][0], bump_range['lhcb2'][0]],
            ele_stop=[bump_range['lhcb1'][-1], bump_range['lhcb2'][-1]],
            twiss_init='preserve',
            targets=targets,
            vary=vary
        )



import json
from pathlib import Path

import xtrack as xt

def machine_tuning(line,
        enable_closed_orbit_correction=False,
        enable_linear_coupling_correction=False,
        enable_tune_correction=False,
        enable_chromaticity_correction=False,
        knob_names=None,
        targets=None,
        line_co_ref=None, co_corr_config=None,
        verbose=False):

    # Correct closed orbit
    if enable_closed_orbit_correction:
        print(f'Correcting closed orbit')
        assert line_co_ref is not None
        assert co_corr_config is not None
        if isinstance(co_corr_config, (str, Path)):
            with open(co_corr_config, 'r') as fid:
                co_corr_config = json.load(fid)

        line.correct_closed_orbit(
                                reference=line_co_ref,
                                correction_config=co_corr_config)

    if enable_linear_coupling_correction:
        assert knob_names is not None
        assert 'c_minus_knob_1' in knob_names
        assert 'c_minus_knob_2' in knob_names
        # Match coupling
        print(f'Matching linear coupling')
        line.match(verbose=verbose,
            vary=[
                xt.Vary(name=knob_names['c_minus_knob_1'],
                        limits=[-0.5e-2, 0.5e-2], step=1e-5),
                xt.Vary(name=knob_names['c_minus_knob_2'],
                        limits=[-0.5e-2, 0.5e-2], step=1e-5)],
            targets=[xt.Target('c_minus', 0, tol=1e-4)])

    # Match tune and chromaticity
    if enable_tune_correction or enable_chromaticity_correction:

        vary = []
        match_targets = []

        if enable_tune_correction:
            assert knob_names is not None
            assert 'q_knob_1' in knob_names
            assert 'q_knob_2' in knob_names
            assert targets is not None
            assert 'qx' in targets
            assert 'qy' in targets

            vary.append(xt.Vary(knob_names['q_knob_1'], step=1e-5))
            vary.append(xt.Vary(knob_names['q_knob_2'], step=1e-5))
            match_targets.append(xt.Target('qx', targets['qx'], tol=1e-4))
            match_targets.append(xt.Target('qy', targets['qy'], tol=1e-4))

        if enable_chromaticity_correction:
            assert knob_names is not None
            assert 'dq_knob_1' in knob_names
            assert 'dq_knob_2' in knob_names
            assert targets is not None
            assert 'dqx' in targets
            assert 'dqy' in targets

            vary.append(xt.Vary(knob_names['dq_knob_1'], step=1e-2))
            vary.append(xt.Vary(knob_names['dq_knob_2'], step=1e-2))
            match_targets.append(xt.Target('dqx', targets['dqx'], tol=0.05))
            match_targets.append(xt.Target('dqy', targets['dqy'], tol=0.05))

        print(f'Matching tune and chromaticity')
        line.match(verbose=verbose, vary=vary, targets=match_targets)



from scipy.constants import c as clight


def rename_coupling_knobs_and_coefficients(line, beamn):

    line.vars[f'c_minus_re_b{beamn}'] = 0
    line.vars[f'c_minus_im_b{beamn}'] = 0
    for ii in [1, 2, 3, 4, 5, 6, 7, 8]:
        for jj, nn in zip([1, 2], ['re', 'im']):
            old_name = f'b{ii}{jj}'
            new_name = f'coeff_skew_{ii}{jj}_b{beamn}'

            # Copy value in new variable
            line.vars[new_name] = line.vars[old_name]._value

            # Zero old variable
            line.vars[old_name] = 0

            # Identify controlled circuit
            targets = line.vars[old_name]._find_dependant_targets()
            if len(targets) > 1: # Controls something
                ttt = [t for t in targets if repr(t).startswith('vars[') and
                    repr(t) != f"vars['{old_name}']"]
                assert len(ttt) > 0
                assert len(ttt) < 3

                for kqs_knob in ttt:
                    kqs_knob_str = repr(kqs_knob)
                    assert "'" in kqs_knob_str
                    assert '"' not in kqs_knob_str
                    var_name = kqs_knob_str.split("'")[1]
                    assert var_name.startswith('kqs')
                    line.vars[var_name] += (line.vars[new_name]
                                * line.vars[f'c_minus_{nn}_b{beamn}'])

def define_octupole_current_knobs(line, beamn):

    line.vars[f'p0c_b{beamn}'] = line.particle_ref.p0c[0]
    line.vars[f'q0_b{beamn}'] = line.particle_ref.q0
    line.vars[f'brho0_b{beamn}'] = (line.vars[f'p0c_b{beamn}']
                                / line.vars[f'q0_b{beamn}'] / clight)

    line.vars[f'i_oct_b{beamn}'] = 0
    for ss in '12 23 34 45 56 67 78 81'.split():
        line.vars[f'kof.a{ss}b{beamn}'] = (
            line.vars['kmax_mo']
            * line.vars[f'i_oct_b{beamn}'] / line.vars['imax_mo']
            / line.vars[f'brho0_b{beamn}'])
        line.vars[f'kod.a{ss}b{beamn}'] = (
            line.vars['kmax_mo']
            * line.vars[f'i_oct_b{beamn}'] / line.vars['imax_mo']
            / line.vars[f'brho0_b{beamn}'])

def add_correction_term_to_dipole_correctors(line):
    # Add correction term to all dipole correctors
    line.vars['on_corr_co'] = 1
    for kk in list(line.vars._owner.keys()):
        if kk.startswith('acb'):
            line.vars['corr_co_'+kk] = 0
            line.vars[kk] += (line.vars['corr_co_'+kk] * line.vars['on_corr_co'])




import yaml
import re

# By default yaml interprets 2.2e11 as string and not as float
# This is a workaround to force yaml to interpret it as float
# See https://github.com/yaml/pyyaml/issues/173
# https://stackoverflow.com/questions/30458977/yaml-loads-5e-6-as-string-and-not-a-number

loader = yaml.SafeLoader
loader.add_implicit_resolver(
    u'tag:yaml.org,2002:float',
    re.compile(u'''^(?:
     [-+]?(?:[0-9][0-9_]*)\\.[0-9_]*(?:[eE][-+]?[0-9]+)?
    |[-+]?(?:[0-9][0-9_]*)(?:[eE][-+]?[0-9]+)
    |\\.[0-9_]+(?:[eE][-+][0-9]+)?
    |[-+]?[0-9][0-9_]*(?::[0-5]?[0-9])+\\.[0-9_]*
    |[-+]?\\.(?:inf|Inf|INF)
    |\\.(?:nan|NaN|NAN))$''', re.X),
    list(u'-+0123456789.'))

def load(data):
    return yaml.safe_load(data)


# copyright ############################### #
# This file is part of the Xtrack Package.  #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

from setuptools import setup, find_packages, Extension
from pathlib import Path

#######################################
# Prepare list of compiled extensions #
#######################################

extensions = []

#########
# Setup #
#########

version_file = Path(__file__).parent / 'xmask/_version.py'
dd = {}
with open(version_file.absolute(), 'r') as fp:
    exec(fp.read(), dd)
__version__ = dd['__version__']

setup(
    name='xmask',
    version=__version__,
    description='Configuration of tracking simulations for the LHC and other accelerators',
    long_description='Configuration of tracking simulations for the LHC and other accelerator',
    url='https://github.com/xsuite/xsuite/issues',
    author='G. Iadarola et al.',
    license='Apache 2.0',
    download_url="https://pypi.python.org/pypi/xmask",
    project_urls={
            "Bug Tracker": "https://github.com/xsuite/xsuite/issues",
            "Source Code": "https://github.com/xsuite/xmask/",
        },
    packages=find_packages(),
    ext_modules = extensions,
    include_package_data=True,
    install_requires=[
        'numpy>=1.0',
        'pyyaml',
        'xfields',
        ],
    extras_require={
        'tests': ['cpymad', 'pytest'],
        },
    )




import os
import pathlib

lhc_module_folder = pathlib.Path(__file__).parent

def install_errors_placeholders_hllhc(mad):
    os.system(f'rm errors')
    os.symlink(lhc_module_folder / 'lhcerrors', 'errors')

    mad.input('''
      ! Install placeholder elements for errors (set to zero)
      call, file="errors/HL-LHC/install_MQXF_fringenl.madx";  ! adding fringe place holder
      call, file="errors/HL-LHC/install_MCBXFAB_errors.madx"; !adding D1 corrector placeholders in IR1/5 (for errors)
      call, file="errors/HL-LHC/install_MCBRD_errors.madx";   ! adding D2 corrector placeholders in IR1/5 (for errors)
      call, file="errors/HL-LHC/install_NLC_errors.madx";     !adding non-linear corrector placeholders in IR1/5 (for errors)
    ''')

def install_correct_errors_and_synthesisize_knobs(mad_track, enable_imperfections,
                        enable_knob_synthesis, pars_for_imperfections,
                        ver_lhc_run=None, ver_hllhc_optics=None):

    os.system(f'rm errors')
    os.symlink(lhc_module_folder / 'lhcerrors', 'errors')

    assert ver_lhc_run is not None or ver_hllhc_optics is not None, (
        'Must specify either ver_lhc_run or ver_hllhc_optics')

    if ver_lhc_run is not None:
        assert ver_hllhc_optics is None, (
            'Must specify either ver_lhc_run or ver_hllhc_optics, not both')
        assert type(ver_lhc_run) is float
        mad_track.globals.ver_lhc_run = ver_lhc_run

    if ver_hllhc_optics is not None:
        assert ver_lhc_run is None, (
            'Must specify either ver_lhc_run or ver_hllhc_optics, not both')
        assert type(ver_hllhc_optics) is float
        mad_track.globals.ver_hllhc_optics = ver_hllhc_optics

    mad_track.globals.par_verbose = 1

    # Force on_disp = 0
    mad_track.globals.on_disp = 0. # will be restored later

    scripts_folder = os.path.dirname(__file__) + '/madx_scripts'

    # Install and correct errors
    if enable_imperfections:
        for kk, vv in pars_for_imperfections.items():
            mad_track.globals[kk] = vv
        mad_track.input(f'call, file="{scripts_folder}/submodule_04a_preparation.madx";')
        mad_track.input(f'call, file="{scripts_folder}/submodule_04a_s1_prepare_nom_twiss_table.madx";')

        mad_track.input('exec, crossing_disable;')

        mad_track.input(f'call, file="{scripts_folder}/submodule_04b_alignsep.madx";')
        mad_track.input(f'call, file="{scripts_folder}/submodule_04c_errortables.madx";')
        mad_track.input(f'call, file="{scripts_folder}/submodule_04d_efcomp.madx";')
        mad_track.input(f'call, file="{scripts_folder}/submodule_04e_s1_synthesize_knobs.madx";')
        mad_track.input(f'call, file="{scripts_folder}/submodule_04e_correction.madx";')
        mad_track.input(f'call, file="{scripts_folder}/submodule_04f_final.madx";')

        mad_track.input('exec, crossing_restore;')
    else:
        # Mock for testing on mac
        if enable_knob_synthesis == '_mock_for_testing':
            if mad_track.globals['mylhcbeam'] == 1:
                mad_track.call("acc-models-lhc/"
                        "_for_test_cminus_knobs_15cm/MB_corr_setting_b1.mad")
            else:
                mad_track.call("acc-models-lhc/"
                        "_for_test_cminus_knobs_15cm/MB_corr_setting_b2.mad")
        elif enable_knob_synthesis:
            mad_track.input(f'call, file="{scripts_folder}/submodule_04a_s1_prepare_nom_twiss_table.madx";')
            mad_track.input(f'call, file="{scripts_folder}/submodule_04e_s1_synthesize_knobs.madx";')



from .errors import install_correct_errors_and_synthesisize_knobs
from .errors import install_errors_placeholders_hllhc
from .knob_manipulations import rename_coupling_knobs_and_coefficients
from .knob_manipulations import define_octupole_current_knobs
from .knob_manipulations import add_correction_term_to_dipole_correctors
from .build_madx_and_xsuite_models import build_xsuite_collider
from .leveling import luminosity_leveling


import xtrack as xt
import xpart as xp

import xmask as xm

from .errors import install_correct_errors_and_synthesisize_knobs
from .knob_manipulations import rename_coupling_knobs_and_coefficients
from .knob_manipulations import define_octupole_current_knobs
from .knob_manipulations import add_correction_term_to_dipole_correctors


# Assumptions from the madx model and optics
# - Machine energy is stored in madx variable "nrj" (in GeV)
# - The variable "mylhcbeam" is set to 1 in the madx instance holding
#   b1 and b2 (both clockwise), and to 4 in the madx instance holding b4
#   (anti-clockwise).
# - Version of optics is stored in madx variable "ver_lhc_run" for LHC and
#   "ver_hllhc_optics" for HL-LHC
# - Macros are available to save/desable/load orbit bumps, which are called
#   "crossing_save", "crossing_disable", and "crossing_restore".


def build_xsuite_collider(
    sequence_b1, sequence_b2, sequence_b4, beam_config,
    enable_imperfections,
    install_apertures=False,
    enable_knob_synthesis=False,
    rename_coupling_knobs=False,
    pars_for_imperfections={},
    ver_lhc_run=None,
    ver_hllhc_optics=None,
    call_after_last_use=None,):

    """
    Build xsuite collider from madx sequences and optics.

    Parameters
    ----------
    sequence_b1: cpymad.madx.Sequence
        Madx sequence for beam 1 (clockwise)
    sequence_b2: cpymad.madx.Sequence
        Madx sequence for beam 2 (clockwise)
    sequence_b4: cpymad.madx.Sequence
        Madx sequence for beam 4 (anti-clockwise)
    beam_config: dict
        Dictionary with beam configuration (see examples)
    enable_imperfections: bool
        If True, lattice imperfections are installed and corrected
    install_apertures: bool
        If True, apertures are installed
    enable_knob_synthesis: bool
        If True, knobs (linear coupling) are synthesized
    rename_coupling_knobs: bool
        If True, coupling knobs are renamed to avoid clashes between b1 and b2
    pars_for_imperfections: dict
        Dictionary with parameters for imperfections configuration (see examples)
    ver_lhc_run: str
        Version of LHC optics (None if HL-LHC)
    ver_hllhc_optics: str
        Version of HL-LHC optics (None if LHC)
    call_after_last_use: callable
        Function called after last madx use command


    Returns
    -------
    collider: xtrack.Multiline

    """

    if sequence_b4 is not None:
        assert sequence_b2 is not None, 'If b4 provided, b2 must be provided'

    for beam_n, sequence in zip((1, 2), [sequence_b1, sequence_b2]):
        if sequence is None: continue
        xm.attach_beam_to_sequence(sequence, beam_to_configure=beam_n,
                            beam_configuration=beam_config[sequence.name])
        # Warm up (seems I need to twiss for mad to load everything)
        sequence._madx.use(sequence.name)
        sequence._madx.twiss()

    # Store energy in nrj
    if sequence_b4 is not None:
        sequence_b4._madx.globals.nrj = sequence_b2._madx.globals.nrj

    # Generate beam 4
    if sequence_b4 is not None:
        xm.configure_b4_from_b2(sequence_b4=sequence_b4, sequence_b2=sequence_b2)

    # Save lines for closed orbit reference
    lines_co_ref = xm.save_lines_for_closed_orbit_reference(
        sequence_clockwise=sequence_b1,
        sequence_anticlockwise=sequence_b4)

    lines_to_track = {}
    for sequence_to_track in [sequence_b1, sequence_b4]:

        if sequence_to_track is None:
            continue

        sequence_name = sequence_to_track.name
        mad_track = sequence_to_track._madx

        # Final use
        mad_track.use(sequence_name)

        if call_after_last_use is not None:
            call_after_last_use(mad_track)

        # We work exclusively on the flat machine
        mad_track.input('exec, crossing_disable;')
        mad_track.input('exec, crossing_save;') # In this way crossing_restore keeps the flat machine

        # Install and correct errors
        install_correct_errors_and_synthesisize_knobs(mad_track,
            enable_imperfections=enable_imperfections,
            enable_knob_synthesis=enable_knob_synthesis,
            pars_for_imperfections=pars_for_imperfections,
            ver_lhc_run=ver_lhc_run,
            ver_hllhc_optics=ver_hllhc_optics)

        # Prepare xsuite line
        line = xt.Line.from_madx_sequence(
            mad_track.sequence[sequence_name], apply_madx_errors=True,
            deferred_expressions=True,
            install_apertures=install_apertures,
            replace_in_expr={'bv_aux': 'bvaux_'+sequence_name})
        mad_beam = mad_track.sequence[sequence_name].beam
        line.particle_ref = xp.Particles(p0c = mad_beam.pc*1e9,
            q0 = mad_beam.charge, mass0 = mad_beam.mass*1e9)

        # Prepare coupling and octupole knobs
        if rename_coupling_knobs:
            rename_coupling_knobs_and_coefficients(line=line,
                                    beamn=int(sequence_name[-1]))

        lines_to_track[sequence_name] = line

    lines = {}
    lines.update(lines_to_track)
    lines.update(lines_co_ref)

    if 'lhcb1' in lines_to_track:
        lines['lhcb1_co_ref'].particle_ref = lines['lhcb1'].particle_ref.copy()
    if 'lhcb2' in lines_to_track:
        lines['lhcb2_co_ref'].particle_ref = lines['lhcb2'].particle_ref.copy()

    collider = xt.Multiline(lines=lines)

    if 'lhcb1' in lines_to_track:
        define_octupole_current_knobs(line=collider.lhcb1, beamn=1)
    if 'lhcb2' in lines_to_track:
        define_octupole_current_knobs(line=collider.lhcb2, beamn=2)

    add_correction_term_to_dipole_correctors(collider)

    return collider